// drOpenGL.dll
// OpenGL driver for the Doomsday Graphics Library
//
// draw.c : Drawing operations, vertex handling

#include "drOpenGL.h"
#include <stdio.h>
#include <stdlib.h>

/*#define DEBUGMODE
#define DEBUG
#include <dprintf.h>
#include <assert.h>*/

// Vertex flags.
#define VTXF_POS_2		0x1
#define VTXF_POS_3		0x2
#define VTXF_POS_MASK	0x3
#define VTXF_COLOR_3	0x4
#define VTXF_COLOR_4	0x8
#define VTXF_COLOR_MASK	0xc
#define VTXF_TEX		0x10		

typedef struct
{
	char flags;
	float pos[3];
	float color[4];
	float tex[2];
}
vertex_t;


int			primType = DGL_FALSE;	// Set in Begin() if the stack is needed.
vertex_t	*vertexStack = 0;
int			stackSize = 0, stackPos = 0;

vertex_t	currentVertex;		
int			inSequence = DGL_FALSE;





// ------------------------------------------------------------------------

void InitVertexStack()
{
	stackSize = 4096;
	vertexStack = malloc(sizeof(vertex_t) * stackSize);
}

void KillVertexStack()
{
	stackSize = 0;
	free(vertexStack);
}

// API FUNCTIONS ----------------------------------------------------------

void Color3ub(DGLubyte r, DGLubyte g, DGLubyte b)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_3;
	vtx->color[0] = r / (float) 0xff;
	vtx->color[1] = g / (float) 0xff;
	vtx->color[2] = b / (float) 0xff;
}


void Color3ubv(void *data)
{
	vertex_t *vtx = vertexStack + stackPos;
	int i;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_3;
	for(i=0; i<3; i++) vtx->color[i] = ((DGLubyte*)data)[i] / (float) 0xff;
}


void Color4ub(DGLubyte r, DGLubyte g, DGLubyte b, DGLubyte a)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_4;
	vtx->color[0] = r / (float) 0xff;
	vtx->color[1] = g / (float) 0xff;
	vtx->color[2] = b / (float) 0xff;
	vtx->color[3] = a / (float) 0xff;
}


void Color4ubv(void *data)
{
	vertex_t *vtx = vertexStack + stackPos;
	int i;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_4;
	for(i=0; i<4; i++) vtx->color[i] = ((DGLubyte*)data)[i] / (float) 0xff;
}


void Color3f(float r, float g, float b)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_3;
	vtx->color[0] = r;
	vtx->color[1] = g;
	vtx->color[2] = b;
}


void Color3fv(float *data)
{
	vertex_t *vtx = vertexStack + stackPos;
	int i;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_3;
	for(i=0; i<3; i++) vtx->color[i] = data[i];
}


void Color4f(float r, float g, float b, float a)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_4;
	vtx->color[0] = r;
	vtx->color[1] = g;
	vtx->color[2] = b;
	vtx->color[3] = a;
}


void Color4fv(float *data)
{
	vertex_t *vtx = vertexStack + stackPos;
	int i;

	vtx->flags &= ~VTXF_COLOR_MASK;
	vtx->flags |= VTXF_COLOR_4;
	for(i=0; i<4; i++) vtx->color[i] = data[i];
}


void TexCoord2f(float s, float t)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags |= VTXF_TEX;
	vtx->tex[0] = s;
	vtx->tex[1] = t;
}


void TexCoord2fv(float *data)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags |= VTXF_TEX;
	vtx->tex[0] = data[0];
	vtx->tex[1] = data[1];
}


void Begin(int mode)
{
	if(mode == DGL_SEQUENCE)
	{
		inSequence = DGL_TRUE;
		return;
	}

	primType = mode;
	stackPos = 0;
}


void End(void)
{
	int	i;
	vertex_t *vtx;

	if(inSequence)
	{
		inSequence = DGL_FALSE;
		return;
	}

	glBegin(primType==DGL_LINES? GL_LINES 
		: primType==DGL_TRIANGLES? GL_TRIANGLES
		: primType==DGL_TRIANGLE_FAN? GL_TRIANGLE_FAN
		: primType==DGL_TRIANGLE_STRIP? GL_TRIANGLE_STRIP
		: primType==DGL_QUAD_STRIP? GL_QUAD_STRIP
		: GL_QUADS);

	for(i=0, vtx=vertexStack; i<stackPos; i++, vtx++)
	{
		if(vtx->flags & VTXF_COLOR_3)
			glColor3fv(vtx->color);
		else if(vtx->flags & VTXF_COLOR_4)
			glColor4fv(vtx->color);
		if(vtx->flags & VTXF_TEX)
			glTexCoord2fv(vtx->tex);
		if(vtx->flags & VTXF_POS_2)
			glVertex2fv(vtx->pos);
		else 
			glVertex3fv(vtx->pos);
	}

	glEnd();
#ifdef DEBUGMODE
	{
		int code = glGetError();
		dprintf("GL Error: %x", code);
		//assert(code == GL_NO_ERROR);
	}
#endif
	primType = DGL_FALSE;
	stackPos = 0;
	vertexStack[0].flags = 0;
}


static void VtxToStack()
{
	stackPos++;
	if(stackPos == stackSize)
	{
		// We need to allocate more memory.
		stackSize += 1024;
		vertexStack = realloc(vertexStack, sizeof(vertex_t) * stackSize);
	}
	vertexStack[stackPos].flags = 0;
}


void Vertex2f(float x, float y)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_POS_MASK;
	vtx->flags |= VTXF_POS_2;
	vtx->pos[0] = x;
	vtx->pos[1] = y;
	
	VtxToStack();
}


void Vertex2fv(float *data)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_POS_MASK;
	vtx->flags |= VTXF_POS_2;
	vtx->pos[0] = data[0];
	vtx->pos[1] = data[1];
	
	VtxToStack();
}


void Vertex3f(float x, float y, float z)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_POS_MASK;
	vtx->flags |= VTXF_POS_3;
	vtx->pos[0] = x;
	vtx->pos[1] = y;
	vtx->pos[2] = z;
	
	VtxToStack();
}


void Vertex3fv(float *data)
{
	vertex_t *vtx = vertexStack + stackPos;

	vtx->flags &= ~VTXF_POS_MASK;
	vtx->flags |= VTXF_POS_3;
	vtx->pos[0] = data[0];
	vtx->pos[1] = data[1];
	vtx->pos[2] = data[2];
	
	VtxToStack();
}


void Vertices2ftv(int num, gl_ft2vertex_t *data)
{
	for(; num>0; num--, data++)
	{
		TexCoord2fv(data->tex);
		Vertex2fv(data->pos);
	}
}


void Vertices3ftv(int num, gl_ft3vertex_t *data)
{
	for(; num>0; num--, data++)
	{
		TexCoord2fv(data->tex);
		Vertex3fv(data->pos);
	}
}


void Vertices3fctv(int num, gl_fct3vertex_t *data)
{
	for(; num>0; num--, data++)
	{
		Color4fv(data->color);
		TexCoord2fv(data->tex);
		Vertex3fv(data->pos);
	}
}
